@@ -7,7 +7,6 @@ import contextlib
 import fnmatch
 import logging
 import os
-import threading
 import subprocess
 import sys
 import time
@@ -24,9 +23,9 @@ from IPython.core.display import HTML
 from IPython.utils.tokenutil import line_at_cursor, token_at_cursor
 from jupyter_client import manager
 from sos._version import __sos_version__, __version__
-from sos.eval import SoS_eval, SoS_exec
+from sos.eval import SoS_eval
 from sos.syntax import SOS_SECTION_HEADER, SOS_DIRECTIVE
-from sos.utils import WorkflowDict, env, short_repr, load_config_files
+from sos.utils import env, short_repr, load_config_files
 from sos.targets import file_target
 from sos.executor_utils import prepare_env
 
@@ -743,6 +742,8 @@ class SoS_Kernel(IPythonKernel):
                 }[env.verbosity],
                 kernel=self))
         env.logger.print = lambda cell_id, msg, *args: \
+            self.send_response(self.iopub_socket, 'stream', {'name': 'stdout', 'text': msg}) \
+                if self._meta['batch_mode'] else \
             self.send_frontend_msg('print', [cell_id, msg])
         self.controller = None
 
@@ -822,9 +823,8 @@ class SoS_Kernel(IPythonKernel):
         # if comm is never created by frontend, the kernel is in test mode without frontend
         if msg_type in ('display_data', 'stream'):
             if self._meta['use_panel'] is False:
-                if msg_type in ('display_data', 'stream'):
-                    self.send_response(self.iopub_socket, msg_type,
-                                       {} if msg is None else msg)
+                self.send_response(self.iopub_socket, msg_type,
+                                {} if msg is None else msg)
             elif self._meta['use_iopub']:
                 self.send_response(self.iopub_socket, 'transient_display_data',
                                    make_transient_msg(msg_type, msg))
@@ -838,6 +838,11 @@ class SoS_Kernel(IPythonKernel):
                 self.frontend_comm.send(
                     make_transient_msg(msg_type, msg),
                     {'msg_type': 'transient_display_data'})
+            elif self._meta['batch_mode']:
+                env.log_to_file(
+                    'MESSAGE',
+                    f'frontend message of type {msg_type} is sent in batch mode.'
+                )
             else:
                 self.frontend_comm_cache.append([msg_type, msg])
                 env.log_to_file(
@@ -852,6 +857,11 @@ class SoS_Kernel(IPythonKernel):
                 self.frontend_comm_cache = []
             self.frontend_comm.send({} if msg is None else msg,
                                     {'msg_type': msg_type})
+        elif self._meta['batch_mode']:
+            env.log_to_file(
+                'MESSAGE',
+                f'frontend message of type {msg_type} is sent in batch mode.'
+            )
         else:
             # frontend_comm is not ready
             self.frontend_comm_cache.append([msg_type, msg])
@@ -974,7 +984,7 @@ class SoS_Kernel(IPythonKernel):
                 # if somethign goes wrong in the subkernel does not matter
                 env.log_to_file(
                     'MAGIC',
-                    f'Failed to call put_var({items}) from {kinfo.kernel}')
+                    f'Failed to call put_var({items}) from {kinfo.kernel}: {e}')
                 objects = {}
             if isinstance(objects, dict):
                 # returns a SOS dictionary
@@ -1107,7 +1117,7 @@ class SoS_Kernel(IPythonKernel):
                 )
                 return msg['content']
 
-            raise RuntimError(
+            raise RuntimeError(
                 f"is_complete_reply not obtained: {msg['header']['msg_type']} {msg['content']} returned instead"
             )
         except Exception as e:
@@ -1131,7 +1141,7 @@ class SoS_Kernel(IPythonKernel):
                 _, KC = self.kernels[cell_kernel.name]
             except Exception as e:
                 env.log_to_file('KERNEL',
-                                f'Failed to get subkernels {cell_kernel.name}')
+                                f'Failed to get subkernels {cell_kernel.name}: {e}')
                 KC = self.KC
             try:
                 KC.inspect(code, cursor_pos)
@@ -1180,7 +1190,7 @@ class SoS_Kernel(IPythonKernel):
             if msg['header']['msg_type'] == 'complete_reply':
                 return msg['content']
 
-            raise RuntimError(
+            raise RuntimeError(
                 f"complete_reply not obtained: {msg['header']['msg_type']} {msg['content']} returned instead"
             )
         except Exception as e:
@@ -1503,7 +1513,7 @@ Available subkernels:\n{}'''.format(
                         code=code,
                         raw_args=self.options,
                         kernel=self,
-                        run_in_queue=self._workflow_mode == 'nowait')
+                        run_in_queue=self._workflow_mode == 'nowait' and not self._meta['batch_mode'])
                 else:
                     res = execute_scratch_cell(
                         code=code, raw_args=self.options, kernel=self)
@@ -1511,12 +1521,7 @@ Available subkernels:\n{}'''.format(
             except Exception as e:
                 sys.stderr.flush()
                 sys.stdout.flush()
-                # self.send_response(self.iopub_socket, 'display_data',
-                #    {
-                #        'metadata': {},
-                #        'data': { 'text/html': HTML('<hr color="black" width="60%">').data}
-                #    })
-                raise
+                raise e
             except KeyboardInterrupt:
                 self.warn('Keyboard Interrupt\n')
                 return {
@@ -1526,65 +1531,6 @@ Available subkernels:\n{}'''.format(
             finally:
                 sys.stderr.flush()
                 sys.stdout.flush()
-        #
-        if False:
-            input_files = [
-                str(x)
-                for x in env.sos_dict.get('step_input', [])
-                if isinstance(x, file_target)
-            ]
-            output_files = [
-                str(x)
-                for x in env.sos_dict.get('step_output', [])
-                if isinstance(x, file_target)
-            ]
-
-            # use a table to list input and/or output file if exist
-            if output_files and not (hasattr(self, '_no_auto_preview') and
-                                     self._no_auto_preview):
-                title = f'%preview {" ".join(output_files)}'
-                if not self._meta['use_panel']:
-                    self.send_response(
-                        self.iopub_socket, 'display_data', {
-                            'metadata': {},
-                            'data': {
-                                'text/html':
-                                    HTML(f'<div class="sos_hint">{title}</div>'
-                                        ).data
-                            }
-                        })
-
-                if hasattr(self, 'in_sandbox') and self.in_sandbox:
-                    # if in sand box, do not link output to their files because these
-                    # files will be removed soon.
-                    self.send_frontend_msg(
-                        'display_data', {
-                            'metadata': {},
-                            'data': {
-                                'text/html':
-                                    '''<div class="sos_hint"> input: {}<br>output: {}\n</div>'''
-                                    .format(', '.join(x for x in input_files),
-                                            ', '.join(x for x in output_files))
-                            }
-                        })
-                else:
-                    self.send_frontend_msg(
-                        'display_data', {
-                            'metadata': {},
-                            'data': {
-                                'text/html':
-                                    '''<div class="sos_hint"> input: {}<br>output: {}\n</div>'''
-                                    .format(
-                                        ', '.join(
-                                            f'<a target="_blank" href="{x}">{x}</a>'
-                                            for x in input_files),
-                                        ', '.join(
-                                            f'<a target="_blank" href="{x}">{x}</a>'
-                                            for x in output_files))
-                            }
-                        })
-
-                Preview_Magic(self).handle_magic_preview(output_files, "SoS")
 
     def render_result(self, res):
         if not self._meta['render_result']:
@@ -1807,7 +1753,7 @@ Available subkernels:\n{}'''.format(
                                    user_expressions, allow_stdin)
         if self.kernel != 'SoS':
             # handle string interpolation before sending to the underlying kernel
-            if self._meta['cell_id'] != '0':
+            if self._meta['cell_id'] != '0' and not self._meta['batch_mode']:
                 self.send_frontend_msg('cell-kernel',
                                        [self._meta['cell_id'], self.kernel])
             if code is None or not code.strip():
@@ -1828,7 +1774,8 @@ Available subkernels:\n{}'''.format(
             # if the cell starts with comment, and newline, remove it
             lines = code.splitlines()
             empties = [x.startswith('#') or not x.strip() for x in lines]
-            self.send_frontend_msg('cell-kernel',
+            if not self._meta['batch_mode']:
+                self.send_frontend_msg('cell-kernel',
                                    [self._meta['cell_id'], 'SoS'])
             if all(empties):
                 return {
